@@ -426,7 +426,7 @@ enum CompressedFormatExtension {
 
 CompressedFormatExtension texturedCompressedFormat(const QString &source)
 {
-    const QString suffix = QFileInfo(source).suffix();
+    const QString suffix = QFileInfo(source).suffix().toLower();
     if (suffix == QStringLiteral("pkm"))
         return PKM;
     if (suffix == QStringLiteral("dds"))
@@ -907,6 +907,11 @@ void QTextureLoaderPrivate::updateGenerator()
     setDataFunctor(QTextureFromSourceGeneratorPtr::create(q));
 }
 
+/*!
+   \class Qt3DRender::QTextureLoader
+   \inmodule Qt3DRender
+   \brief Handles the texture loading and setting the texture's properties.
+*/
 /*!
  * Constructs a new Qt3DRender::QTextureLoader instance with \a parent as parent.
  *
@@ -943,8 +948,10 @@ QTextureLoader::~QTextureLoader()
 }
 
 /*!
- * \return the current texture source.
- */
+    \property QTextureLoader::source
+
+    Returns the current texture source.
+*/
 QUrl QTextureLoader::source() const
 {
     Q_D(const QTextureLoader);
