@@ -37,10 +37,14 @@
 #include <Qt3DCore/qnodedestroyedchange.h>
 #include <Qt3DCore/qpropertynodeaddedchange.h>
 #include <Qt3DCore/qpropertynoderemovedchange.h>
+#include <Qt3DCore/private/qnodecreatedchangegenerator_p.h>
 #include <private/qpostman_p.h>
 
 #include <Qt3DCore/private/qlockableobserverinterface_p.h>
 #include <Qt3DCore/private/qnode_p.h>
+#include <Qt3DCore/private/qcomponent_p.h>
+#include <QSignalSpy>
+#include "testpostmanarbiter.h"
 
 class tst_Nodes : public QObject
 {
@@ -53,6 +57,7 @@ public:
     ~tst_Nodes() {}
 
 private slots:
+    void initTestCase();
     void defaultNodeConstruction();
     void defaultComponentConstruction();
     void defaultEntityConstrution();
@@ -74,12 +79,21 @@ private slots:
     void appendingChildEntitiesToNode();
     void removingChildEntitiesFromNode();
 
+    void checkConstructionSetParentMix(); // QTBUG-60612
+
     void appendingComponentToEntity();
     void appendingParentlessComponentToEntity();
     void removingComponentFromEntity();
 
     void changeCustomProperty();
     void checkDestruction();
+
+    void checkDefaultConstruction();
+    void checkPropertyChanges();
+    void checkCreationData();
+    void checkEnabledUpdate();
+    void checkPropertyTrackModeUpdate();
+    void checkTrackedPropertyNamesUpdate();
 };
 
 class ObserverSpy;
@@ -93,6 +107,7 @@ public:
     void sceneChangeEvent(const Qt3DCore::QSceneChangePtr &) Q_DECL_FINAL {};
     void setScene(Qt3DCore::QScene *) Q_DECL_FINAL {};
     void notifyBackend(const Qt3DCore::QSceneChangePtr &change) Q_DECL_FINAL;
+    bool shouldNotifyFrontend(const Qt3DCore::QSceneChangePtr &changee) Q_DECL_FINAL { Q_UNUSED(changee); return false; }
 
 private:
     ObserverSpy *m_spy;
@@ -191,6 +206,11 @@ public:
         Qt3DCore::QNodePrivate::get(this)->setArbiter(arbiter);
     }
 
+    void setSimulateBackendCreated(bool created)
+    {
+        Qt3DCore::QNodePrivate::get(this)->m_hasBackendNode = created;
+    }
+
 signals:
     void customPropertyChanged();
 
@@ -219,6 +239,11 @@ public:
         Qt3DCore::QNodePrivate::get(this)->setScene(scene);
         Qt3DCore::QNodePrivate::get(this)->setArbiter(arbiter);
     }
+
+    void setSimulateBackendCreated(bool created)
+    {
+        Qt3DCore::QNodePrivate::get(this)->m_hasBackendNode = created;
+    }
 };
 
 class MyQComponent : public Qt3DCore::QComponent
@@ -227,9 +252,19 @@ class MyQComponent : public Qt3DCore::QComponent
 public:
     explicit MyQComponent(Qt3DCore::QNode *parent = 0) : QComponent(parent)
     {}
+    void setArbiter(Qt3DCore::QAbstractArbiter *arbiter)
+    {
+        Q_ASSERT(arbiter);
+        Qt3DCore::QComponentPrivate::get(this)->setArbiter(arbiter);
+    }
 };
 
 
+void tst_Nodes::initTestCase()
+{
+    qRegisterMetaType<Qt3DCore::QNode::PropertyTrackingMode>("PropertyTrackingMode");
+}
+
 void tst_Nodes::defaultNodeConstruction()
 {
     // GIVEN
@@ -366,6 +401,8 @@ void tst_Nodes::appendSingleChildNodeToNodeSceneExplicitParenting()
     QScopedPointer<MyQNode> node(new MyQNode());
     // WHEN
     node->setArbiterAndScene(&spy, &scene);
+    node->setSimulateBackendCreated(true);
+
     // THEN
     QVERIFY(Qt3DCore::QNodePrivate::get(node.data())->scene() != nullptr);
 
@@ -382,7 +419,7 @@ void tst_Nodes::appendSingleChildNodeToNodeSceneExplicitParenting()
     // THEN
     QVERIFY(child->parent() == node.data());
     QVERIFY(child->parentNode() == node.data());
-    QCOMPARE(spy.events.size(), 2);
+    QCOMPARE(spy.events.size(), 2); // Created + Child Added
     QCOMPARE(node->children().count(), 1);
     QVERIFY(Qt3DCore::QNodePrivate::get(child.data())->scene() != nullptr);
 
@@ -459,6 +496,7 @@ void tst_Nodes::appendMultipleChildNodesToNodeScene()
 
     // WHEN
     node->setArbiterAndScene(&spy, &scene);
+    node->setSimulateBackendCreated(true);
     // THEN
     QVERIFY(Qt3DCore::QNodePrivate::get(node.data())->scene() != nullptr);
 
@@ -641,6 +679,7 @@ void tst_Nodes::removingSingleChildNodeFromNode()
 
     // WHEN
     root->setArbiterAndScene(&spy, &scene);
+    root->setSimulateBackendCreated(true);
     child->setParent(root.data());
 
     // Clear any creation event
@@ -767,109 +806,175 @@ void tst_Nodes::removingChildEntitiesFromNode()
     QVERIFY(childEntity->parent() == nullptr);
 }
 
-void tst_Nodes::appendingParentlessComponentToEntity()
+void tst_Nodes::checkConstructionSetParentMix()
 {
     // GIVEN
     ObserverSpy spy;
-    QScopedPointer<MyQEntity> entity(new MyQEntity());
-    entity->setArbiterAndScene(&spy);
-    MyQComponent *comp = new MyQComponent();
-
-    // THEN
-    QVERIFY(entity->parentNode() == nullptr);
-    QVERIFY(entity->children().count() == 0);
-    QVERIFY(entity->components().empty());
-    QVERIFY(comp->parentNode() == nullptr);
+    Qt3DCore::QScene scene;
+    QScopedPointer<MyQNode> root(new MyQNode());
 
     // WHEN
-    entity->addComponent(comp);
+    root->setArbiterAndScene(&spy, &scene);
+    root->setSimulateBackendCreated(true);
 
     // THEN
-    QVERIFY(entity->components().count() == 1);
-    QVERIFY(entity->components().first() == comp);
-    QVERIFY(comp->parentNode() == entity.data());
-    QCOMPARE(spy.events.size(), 1);
-    QVERIFY(spy.events.first().wasLocked());
+    QVERIFY(Qt3DCore::QNodePrivate::get(root.data())->scene() != nullptr);
 
-    // Note: since QEntity has no scene in this test case, we only have the
-    // ComponentAdded event In theory we should also get the NodeCreated event
-    // when setting the parent but that doesn't happen since no scene is
-    // actually set on the entity and that QNodePrivate::_q_addChild will
-    // return early in such a case.
+    // WHEN
+    Qt3DCore::QEntity *subTreeRoot = new Qt3DCore::QEntity(root.data());
 
-    // Check that we received ComponentAdded
-    const auto event = spy.events.takeFirst().change().dynamicCast<Qt3DCore::QComponentAddedChange>();
-    QCOMPARE(event->type(), Qt3DCore::ComponentAdded);
-    QCOMPARE(event->subjectId(), entity->id());
-    QCOMPARE(event->entityId(), entity->id());
-    QCOMPARE(event->componentId(), comp->id());
-    QCOMPARE(event->componentMetaObject(), comp->metaObject());
+    for (int i = 0; i < 100; ++i) {
+        Qt3DCore::QEntity *child = new Qt3DCore::QEntity();
+        child->setParent(subTreeRoot);
+    }
+
+    // THEN
+    QCoreApplication::processEvents();
+    QCOMPARE(root->children().count(), 1);
+    QCOMPARE(subTreeRoot->children().count(), 100);
+    QCOMPARE(spy.events.size(), 102); // 1 subTreeRoot creation change, 100 child creation, 1 child added (subTree to root)
+
+    // Ensure first event is subTreeRoot
+    const Qt3DCore::QNodeCreatedChangeBasePtr firstEvent = spy.events.takeFirst().change().dynamicCast<Qt3DCore::QNodeCreatedChangeBase>();
+    QVERIFY(!firstEvent.isNull());
+    QCOMPARE(firstEvent->subjectId(), subTreeRoot->id());
+    QCOMPARE(firstEvent->parentId(), root->id());
+
+    const Qt3DCore::QPropertyNodeAddedChangePtr lastEvent = spy.events.takeLast().change().dynamicCast<Qt3DCore::QPropertyNodeAddedChange>();
+    QVERIFY(!lastEvent.isNull());
+    QCOMPARE(lastEvent->subjectId(), root->id());
+    QCOMPARE(lastEvent->propertyName(), "children");
+    QCOMPARE(lastEvent->addedNodeId(), subTreeRoot->id());
+}
+
+void tst_Nodes::appendingParentlessComponentToEntity()
+{
+    // GIVEN
+    ObserverSpy entitySpy;
+    ObserverSpy componentSpy;
+    {
+        QScopedPointer<MyQEntity> entity(new MyQEntity());
+        entity->setArbiterAndScene(&entitySpy);
+        entity->setSimulateBackendCreated(true);
+
+        MyQComponent *comp = new MyQComponent();
+        comp->setArbiter(&componentSpy);
+
+        // THEN
+        QVERIFY(entity->parentNode() == nullptr);
+        QVERIFY(entity->children().count() == 0);
+        QVERIFY(entity->components().empty());
+        QVERIFY(comp->parentNode() == nullptr);
+
+        // WHEN
+        entity->addComponent(comp);
+
+        // THEN
+        QVERIFY(entity->components().count() == 1);
+        QVERIFY(entity->components().first() == comp);
+        QVERIFY(comp->parentNode() == entity.data());
+        QCOMPARE(entitySpy.events.size(), 1);
+        QVERIFY(entitySpy.events.first().wasLocked());
+        QCOMPARE(componentSpy.events.size(), 1);
+
+        // Note: since QEntity has no scene in this test case, we only have the
+        // ComponentAdded event In theory we should also get the NodeCreated event
+        // when setting the parent but that doesn't happen since no scene is
+        // actually set on the entity and that QNodePrivate::_q_addChild will
+        // return early in such a case.
+
+        // Check that we received ComponentAdded
+        for (const auto event: { entitySpy.events.takeFirst().change().dynamicCast<Qt3DCore::QComponentAddedChange>(),
+                                 componentSpy.events.takeLast().change().dynamicCast<Qt3DCore::QComponentAddedChange>() })
+        {
+            QCOMPARE(event->type(), Qt3DCore::ComponentAdded);
+            QCOMPARE(event->subjectId(), entity->id());
+            QCOMPARE(event->entityId(), entity->id());
+            QCOMPARE(event->componentId(), comp->id());
+            QCOMPARE(event->componentMetaObject(), comp->metaObject());
+        }
+    }
 }
 
 void tst_Nodes::appendingComponentToEntity()
 {
     // GIVEN
-    ObserverSpy spy;
-    QScopedPointer<MyQEntity> entity(new MyQEntity());
-    entity->setArbiterAndScene(&spy);
-    MyQComponent *comp = new MyQComponent(entity.data());
-    QCoreApplication::processEvents();
+    ObserverSpy entitySpy;
+    ObserverSpy componentSpy;
+    {
+        QScopedPointer<MyQEntity> entity(new MyQEntity());
+        entity->setArbiterAndScene(&entitySpy);
+        MyQComponent *comp = new MyQComponent(entity.data());
+        comp->setArbiter(&componentSpy);
+        QCoreApplication::processEvents();
 
-    // THEN
-    QVERIFY(entity->parentNode() == nullptr);
-    QVERIFY(entity->children().count() == 1);
-    QVERIFY(entity->components().empty());
-    QVERIFY(comp->parentNode() == entity.data());
+        // THEN
+        QVERIFY(entity->parentNode() == nullptr);
+        QVERIFY(entity->children().count() == 1);
+        QVERIFY(entity->components().empty());
+        QVERIFY(comp->parentNode() == entity.data());
 
-    // WHEN
-    entity->addComponent(comp);
+        // WHEN
+        entity->addComponent(comp);
 
-    // THEN
-    QVERIFY(entity->components().count() == 1);
-    QVERIFY(entity->components().first() == comp);
-    QVERIFY(comp->parentNode() == entity.data());
-    QCOMPARE(spy.events.size(), 1);
-    QVERIFY(spy.events.first().wasLocked());
-    const auto event = spy.events.takeFirst().change().dynamicCast<Qt3DCore::QComponentAddedChange>();
-    QCOMPARE(event->type(), Qt3DCore::ComponentAdded);
-    QCOMPARE(event->subjectId(), entity->id());
-    QCOMPARE(event->entityId(), entity->id());
-    QCOMPARE(event->componentId(), comp->id());
-    QCOMPARE(event->componentMetaObject(), comp->metaObject());
+        // THEN
+        QVERIFY(entity->components().count() == 1);
+        QVERIFY(entity->components().first() == comp);
+        QVERIFY(comp->parentNode() == entity.data());
+        QCOMPARE(entitySpy.events.size(), 1);
+        QVERIFY(entitySpy.events.first().wasLocked());
+        for (const auto event: { entitySpy.events.takeFirst().change().dynamicCast<Qt3DCore::QComponentAddedChange>(),
+             componentSpy.events.takeFirst().change().dynamicCast<Qt3DCore::QComponentAddedChange>() })
+        {
+            QCOMPARE(event->type(), Qt3DCore::ComponentAdded);
+            QCOMPARE(event->subjectId(), entity->id());
+            QCOMPARE(event->entityId(), entity->id());
+            QCOMPARE(event->componentId(), comp->id());
+            QCOMPARE(event->componentMetaObject(), comp->metaObject());
+        }
+    }
 }
 
 void tst_Nodes::removingComponentFromEntity()
 {
     // GIVEN
-    ObserverSpy spy;
-    QScopedPointer<MyQEntity> entity(new MyQEntity());
-    entity->setArbiterAndScene(&spy);
-    MyQComponent *comp = new MyQComponent();
+    ObserverSpy entitySpy;
+    ObserverSpy componentSpy;
+    {
+        QScopedPointer<MyQEntity> entity(new MyQEntity());
+        entity->setArbiterAndScene(&entitySpy);
+        MyQComponent *comp = new MyQComponent();
+        comp->setArbiter(&componentSpy);
 
-    // WHEN
-    entity->addComponent(comp);
+        // WHEN
+        entity->addComponent(comp);
 
-    // THEN
-    QVERIFY(entity->components().count() == 1);
-    QCOMPARE(entity->children().count(), 1);
-    QVERIFY(comp->parent() == entity.data());
+        // THEN
+        QVERIFY(entity->components().count() == 1);
+        QCOMPARE(entity->children().count(), 1);
+        QVERIFY(comp->parent() == entity.data());
 
-    // WHEN
-    spy.events.clear();
-    entity->removeComponent(comp);
+        // WHEN
+        entitySpy.events.clear();
+        componentSpy.events.clear();
+        entity->removeComponent(comp);
 
-    // THEN
-    QVERIFY(entity->components().count() == 0);
-    QVERIFY(comp->parent() == entity.data());
-    QVERIFY(entity->children().count() == 1);
-    QCOMPARE(spy.events.size(), 1);
-    QVERIFY(spy.events.first().wasLocked());
-    const auto event = spy.events.takeFirst().change().dynamicCast<Qt3DCore::QComponentRemovedChange>();
-    QCOMPARE(event->type(), Qt3DCore::ComponentRemoved);
-    QCOMPARE(event->subjectId(), entity->id());
-    QCOMPARE(event->entityId(), entity->id());
-    QCOMPARE(event->componentId(), comp->id());
-    QCOMPARE(event->componentMetaObject(), comp->metaObject());
+        // THEN
+        QVERIFY(entity->components().count() == 0);
+        QVERIFY(comp->parent() == entity.data());
+        QVERIFY(entity->children().count() == 1);
+        QCOMPARE(entitySpy.events.size(), 1);
+        QVERIFY(entitySpy.events.first().wasLocked());
+        QCOMPARE(componentSpy.events.size(), 1);
+        for (const auto event: { entitySpy.events.takeFirst().change().dynamicCast<Qt3DCore::QComponentRemovedChange>(),
+                                 componentSpy.events.takeFirst().change().dynamicCast<Qt3DCore::QComponentRemovedChange>() }) {
+            QCOMPARE(event->type(), Qt3DCore::ComponentRemoved);
+            QCOMPARE(event->subjectId(), entity->id());
+            QCOMPARE(event->entityId(), entity->id());
+            QCOMPARE(event->componentId(), comp->id());
+            QCOMPARE(event->componentMetaObject(), comp->metaObject());
+        }
+    }
 }
 
 void tst_Nodes::changeCustomProperty()
@@ -915,6 +1020,237 @@ void tst_Nodes::checkDestruction()
     QVERIFY(root->children().isEmpty());
 }
 
+void tst_Nodes::checkDefaultConstruction()
+{
+    // GIVEN
+    Qt3DCore::QNode node;
+
+    // THEN
+    QCOMPARE(node.parentNode(), nullptr);
+    QCOMPARE(node.isEnabled(), true);
+    QCOMPARE(node.defaultPropertyTrackingMode(), Qt3DCore::QNode::TrackFinalValues);
+}
+
+void tst_Nodes::checkPropertyChanges()
+{
+    // GIVEN
+    Qt3DCore::QNode parentNode;
+    Qt3DCore::QNode node;
+
+    {
+        // WHEN
+        QSignalSpy spy(&node, SIGNAL(parentChanged(QObject *)));
+        Qt3DCore::QNode *newValue = &parentNode;
+        node.setParent(newValue);
+
+        // THEN
+        QVERIFY(spy.isValid());
+        QCOMPARE(node.parentNode(), newValue);
+        QCOMPARE(spy.count(), 1);
+
+        // WHEN
+        spy.clear();
+        node.setParent(newValue);
+
+        // THEN
+        QCOMPARE(node.parentNode(), newValue);
+        QCOMPARE(spy.count(), 0);
+    }
+    {
+        // WHEN
+        QSignalSpy spy(&node, SIGNAL(enabledChanged(bool)));
+        const bool newValue = false;
+        node.setEnabled(newValue);
+
+        // THEN
+        QVERIFY(spy.isValid());
+        QCOMPARE(node.isEnabled(), newValue);
+        QCOMPARE(spy.count(), 1);
+
+        // WHEN
+        spy.clear();
+        node.setEnabled(newValue);
+
+        // THEN
+        QCOMPARE(node.isEnabled(), newValue);
+        QCOMPARE(spy.count(), 0);
+    }
+    {
+        // WHEN
+        QSignalSpy spy(&node, SIGNAL(defaultPropertyTrackingModeChanged(PropertyTrackingMode)));
+        const Qt3DCore::QNode::PropertyTrackingMode newValue = Qt3DCore::QNode::TrackAllValues;
+        node.setDefaultPropertyTrackingMode(newValue);
+
+        // THEN
+        QVERIFY(spy.isValid());
+        QCOMPARE(node.defaultPropertyTrackingMode(), newValue);
+        QCOMPARE(spy.count(), 1);
+
+        // WHEN
+        spy.clear();
+        node.setDefaultPropertyTrackingMode(newValue);
+
+        // THEN
+        QCOMPARE(node.defaultPropertyTrackingMode(), newValue);
+        QCOMPARE(spy.count(), 0);
+    }
+    {
+        // WHEN
+        const QString enabledPropertyName = QStringLiteral("enabled");
+        node.setDefaultPropertyTrackingMode(Qt3DCore::QNode::DontTrackValues);
+        node.setPropertyTracking(enabledPropertyName, Qt3DCore::QNode::TrackAllValues);
+
+        // THEN
+        QCOMPARE(node.propertyTracking(enabledPropertyName), Qt3DCore::QNode::TrackAllValues);
+
+        // WHEN
+        node.clearPropertyTracking(enabledPropertyName);
+
+        // THEN
+        QCOMPARE(node.propertyTracking(enabledPropertyName), Qt3DCore::QNode::DontTrackValues);
+    }
+}
+
+void tst_Nodes::checkCreationData()
+{
+    // GIVEN
+    Qt3DCore::QNode root;
+    Qt3DCore::QNode node;
+
+    node.setParent(&root);
+    node.setEnabled(true);
+    const QString enabledPropertyName = QStringLiteral("enabled");
+    node.setDefaultPropertyTrackingMode(Qt3DCore::QNode::DontTrackValues);
+    node.setPropertyTracking(enabledPropertyName, Qt3DCore::QNode::TrackAllValues);
+
+    // WHEN
+    QVector<Qt3DCore::QNodeCreatedChangeBasePtr> creationChanges;
+
+    {
+        Qt3DCore::QNodeCreatedChangeGenerator creationChangeGenerator(&node);
+        creationChanges = creationChangeGenerator.creationChanges();
+    }
+
+    // THEN
+    {
+        QCOMPARE(creationChanges.size(), 1);
+
+        const auto creationChangeData = qSharedPointerCast<Qt3DCore::QNodeCreatedChangeBase>(creationChanges.first());
+
+        QCOMPARE(node.id(), creationChangeData->subjectId());
+        QCOMPARE(node.isEnabled(), true);
+        QCOMPARE(node.isEnabled(), creationChangeData->isNodeEnabled());
+        QCOMPARE(node.metaObject(), creationChangeData->metaObject());
+    }
+
+    // WHEN
+    node.setEnabled(false);
+
+    {
+        Qt3DCore::QNodeCreatedChangeGenerator creationChangeGenerator(&node);
+        creationChanges = creationChangeGenerator.creationChanges();
+    }
+
+    // THEN
+    {
+        QCOMPARE(creationChanges.size(), 1);
+
+        const auto creationChangeData = qSharedPointerCast<Qt3DCore::QNodeCreatedChangeBase>(creationChanges.first());
+
+        QCOMPARE(node.id(), creationChangeData->subjectId());
+        QCOMPARE(node.isEnabled(), false);
+        QCOMPARE(node.isEnabled(), creationChangeData->isNodeEnabled());
+        QCOMPARE(node.metaObject(), creationChangeData->metaObject());
+    }
+}
+
+void tst_Nodes::checkEnabledUpdate()
+{
+    // GIVEN
+    TestArbiter arbiter;
+    Qt3DCore::QNode node;
+    arbiter.setArbiterOnNode(&node);
+
+    {
+        // WHEN
+        node.setEnabled(false);
+        QCoreApplication::processEvents();
+
+        // THEN
+        QCOMPARE(arbiter.events.size(), 1);
+        auto change = arbiter.events.first().staticCast<Qt3DCore::QPropertyUpdatedChange>();
+        QCOMPARE(change->propertyName(), "enabled");
+        QCOMPARE(change->value().value<bool>(), node.isEnabled());
+        QCOMPARE(change->type(), Qt3DCore::PropertyUpdated);
+
+        arbiter.events.clear();
+    }
+
+    {
+        // WHEN
+        node.setEnabled(false);
+        QCoreApplication::processEvents();
+
+        // THEN
+        QCOMPARE(arbiter.events.size(), 0);
+    }
+
+}
+
+void tst_Nodes::checkPropertyTrackModeUpdate()
+{
+    // GIVEN
+    TestArbiter arbiter;
+    Qt3DCore::QNode node;
+    arbiter.setArbiterOnNode(&node);
+
+    {
+        // WHEN
+        node.setDefaultPropertyTrackingMode(Qt3DCore::QNode::TrackAllValues);
+        QCoreApplication::processEvents();
+
+        // THEN -> this properties is non notifying
+        QCOMPARE(arbiter.events.size(), 0);
+    }
+
+    {
+        // WHEN
+        node.setDefaultPropertyTrackingMode(Qt3DCore::QNode::TrackAllValues);
+        QCoreApplication::processEvents();
+
+        // THEN
+        QCOMPARE(arbiter.events.size(), 0);
+    }
+
+}
+
+void tst_Nodes::checkTrackedPropertyNamesUpdate()
+{
+    // GIVEN
+    TestArbiter arbiter;
+    Qt3DCore::QNode node;
+    arbiter.setArbiterOnNode(&node);
+
+    {
+        // WHEN
+        node.setPropertyTracking(QStringLiteral("883"), Qt3DCore::QNode::TrackAllValues);
+        QCoreApplication::processEvents();
+
+        // THEN -> this properties is non notifying
+        QCOMPARE(arbiter.events.size(), 0);
+    }
+
+    {
+        // WHEN
+        node.setPropertyTracking(QStringLiteral("883"), Qt3DCore::QNode::DontTrackValues);
+        QCoreApplication::processEvents();
+
+        // THEN
+        QCOMPARE(arbiter.events.size(), 0);
+    }
+
+}
+
 QTEST_MAIN(tst_Nodes)
 
 #include "tst_nodes.moc"
